/**
 * This file is in the public domain.
 * You may do whatever you want with it.
 */
package org.finance.data;

import org.calendar.SimpleDate;

import java.util.*;
import org.prevayler.*;
import java.io.*;
import org.prevaylerwrapper.*;

public class FinancesImpl extends DefaultCommandSystem implements Finances {

  // long day   -> Hashtable {String name -> long value}
  private static String dailyExpenses = "dailyExpenses";

  // long day   -> Hashtable {String name -> long value}
  private static String cardIncome = "cardIncome";

  // long day   -> Hashtable {String name -> long value}
  private static String cardWithdraw = "cardWithdraw";

  // long month -> Hashtable {String name -> long value}
  private static String monthlyIncome = "monthlyIncome";

  // long month -> Hashtable {String name -> long value}
  private static String plannedMontlyExpenses = "plannedMontlyExpenses";

  // long month -> long value
  private static String plannedEveryDayExpenses = "plannedEveryDayExpenses";

  // long day   -> Hashtable {String name -> long value}
  private static String plannedWeeklyExpenses = "plannedWeeklyExpenses";

  // long day   -> DailyPlanning
  private static String plannedDailyExpenses = "plannedDailyExpenses";

  // data conversion methods:

  private Object[] getSimpleData(Payment payment) {
    return new Object[] {payment.getName(), new Long(payment.getPureValue())};
  }

  private void addDailyValue(Payment payment, Hashtable data) {
    SimpleDate date = payment.getDate();
    Long dateLong = new Long(date.toLong());
    Hashtable table = (Hashtable)data.get(dateLong);
    if (table == null) {
      table = new Hashtable();
      data.put(dateLong, table);
    }
    table.remove(payment.getName());
    table.put(payment.getName(), new Long(payment.getPureValue()));
  }

  private void removeDailyValue(Payment payment, Hashtable data) {
    SimpleDate date = payment.getDate();
    Long dateLong = new Long(date.toLong());
    Hashtable table = (Hashtable)data.get(dateLong);
    if (table != null) {
      table.remove(payment.getName());
      if (table.size() < 1) {
        data.remove(dateLong);
      }
    }
  }

  private void addMonthlyValue(Payment payment, Hashtable data) {
    SimpleDate date = payment.getDate();
    SimpleDate month = date.getMonthAsDate();
    Long monthLong = new Long(month.toLong());
    Hashtable table = (Hashtable)data.get(monthLong);
    if (table == null) {
      table = new Hashtable();
      data.put(monthLong, table);
    }
    table.remove(payment.getName());
    table.put(payment.getName(), new Long(payment.getPureValue()));
  }

  private void removeMonthlyValue(Payment payment, Hashtable data) {
    SimpleDate date = payment.getDate();
    SimpleDate month = date.getMonthAsDate();
    Long monthLong = new Long(month.toLong());
    Hashtable table = (Hashtable)data.get(monthLong);
    if (table != null) {
      table.remove(payment.getName());
      if (table.size() < 1) {
        data.remove(monthLong);
      }
    }
  }

  private long sumHashtable(Hashtable table) {
    long result = 0;
    if (table != null) {
      for (Enumeration elements = table.elements(); elements.hasMoreElements();) {
        Long next = (Long)elements.nextElement();
        result += next.longValue();
      }
    }
    return result;
  }

  private long sumMonthlyValues(SimpleDate month, Hashtable data) {
    Long monthLong = new Long(month.getMonthAsDate().toLong());
    Hashtable table = (Hashtable)data.get(monthLong);
    return sumHashtable(table);
  }

  private long sumDailyValues(SimpleDate month, SimpleDate to, Hashtable data) {
    month = month.getMonthAsDate();
    SimpleDate nextMonth = to != null ? to : month.getNextMonth();
    long result = 0;
    long begin = month.toLong();
    long afterEnd = nextMonth.toLong();
    if (to != null) {
      if (afterEnd >= Long.MAX_VALUE - 10) {
        throw new RuntimeException("Date too high!");
      }
      afterEnd++;
    }
    for (long i = begin; i < afterEnd; i++) {
      Hashtable table = (Hashtable)data.get(new Long(i));
      result += sumHashtable(table);
    }
    return result;
  }

  private void collectHashtable(SimpleDate date, Vector result, Hashtable table) {
    if (table != null) {
      for (Enumeration keys = table.keys(); keys.hasMoreElements();) {
        String nextName = (String)keys.nextElement();
        Long nextValue = (Long)table.get(nextName);
        Payment nextPayment = new Payment(
          date,
          nextName,
          nextValue.longValue()
        );
        result.addElement(nextPayment);
      }
    }
  }

  private Payment[] collectMonthly(SimpleDate month, Hashtable data) {
    Vector result = new Vector();
    month = month.getMonthAsDate();
    Long monthLong = new Long(month.toLong());
    Hashtable table = (Hashtable)data.get(monthLong);
    collectHashtable(month, result, table);
    return ret(result);
  }

  private Payment[] collectDaily(SimpleDate month, Hashtable data) {
    Vector result = new Vector();
    month = month.getMonthAsDate();
    SimpleDate nextMonth = month.getNextMonth();
    long begin = month.toLong();
    long afterEnd = nextMonth.toLong();
    for (long i = begin; i < afterEnd; i++) {
      Hashtable table = (Hashtable)data.get(new Long(i));
      collectHashtable(new SimpleDate(i), result, table);
    }
    return ret(result);
  }

  private Payment[] collectDailySummaries(SimpleDate month, Hashtable data) {
    month = month.getMonthAsDate();
    SimpleDate nextMonth = month.getNextMonth();
    return collectDailySummaries(month, nextMonth, data);
  }

  private Payment[] collectDailySummaries(
    SimpleDate beginDate,
    SimpleDate afterEndDate,
    Hashtable data
  ) {
    Vector result = new Vector();
    long begin = beginDate.toLong();
    long afterEnd = afterEndDate.toLong();
    for (long i = begin; i < afterEnd; i++) {
      Hashtable table = (Hashtable)data.get(new Long(i));
      long sum = sumHashtable(table);
      result.addElement(new Payment(new SimpleDate(i), "", sum));
    }
    return ret(result);
  }

  // mutators:

  private void addDailyExpense(Payment payment) {
    addDailyValue(payment, getHashtable(dailyExpenses));
  }

  public void updateDailyExpense(Payment oldPayment, Payment newPayment) {
    SimpleDate newDate = newPayment.getDate();
    String newName = newPayment.getName();
    long newValue = newPayment.getPureValue();
    SimpleDate oldDate = oldPayment.getDate();
    String oldName = oldPayment.getName();
    long oldValue = oldPayment.getPureValue();
    boolean dateDiff = !newDate.equals(oldDate);
    boolean nameDiff = !newName.equals(oldName);
    if (dateDiff || nameDiff) {
      removeDailyExpense(oldPayment);
    }
    if (dateDiff || nameDiff || (newValue != oldValue)) {
      addDailyExpense(newPayment);
    }
  }

  public void removeDailyExpense(Payment payment) {
    removeDailyValue(payment, getHashtable(dailyExpenses));
  }

  private void addMonthlyIncome(Payment payment) {
    addMonthlyValue(payment, getHashtable(monthlyIncome));
  }

  public void updateMonthlyIncome(Payment oldPayment, Payment newPayment) {
    String newName = newPayment.getName();
    long newValue = newPayment.getPureValue();
    String oldName = oldPayment.getName();
    long oldValue = oldPayment.getPureValue();
    boolean nameDiff = !newName.equals(oldName);
    if (nameDiff) {
      removeMonthlyIncome(oldPayment);
    }
    if (nameDiff || (newValue != oldValue)) {
      addMonthlyIncome(newPayment);
    }
  }

  public void removeMonthlyIncome(Payment payment) {
    removeMonthlyValue(payment, getHashtable(monthlyIncome));
  }

  private void addPlannedMonthlyExpense(Payment payment) {
    addMonthlyValue(payment, getHashtable(plannedMontlyExpenses));
  }

  public void updatePlannedMonthlyExpense(Payment oldPayment, Payment newPayment) {
    String newName = newPayment.getName();
    long newValue = newPayment.getPureValue();
    String oldName = oldPayment.getName();
    long oldValue = oldPayment.getPureValue();
    boolean nameDiff = !newName.equals(oldName);
    if (nameDiff) {
      removePlannedMonthlyExpense(oldPayment);
    }
    if (nameDiff || (newValue != oldValue)) {
      addPlannedMonthlyExpense(newPayment);
    }
  }

  public void removePlannedMonthlyExpense(Payment payment) {
    removeMonthlyValue(payment, getHashtable(plannedMontlyExpenses));
  }

  public void setPlannedEveryDayExpense(Payment payment) {
    Long monthLong = new Long(payment.getDate().getMonthAsDate().toLong());
    Hashtable plannedEveryDayExpenses = getHashtable(this.plannedEveryDayExpenses);
    plannedEveryDayExpenses.remove(monthLong);
    plannedEveryDayExpenses.put(monthLong, new Long(payment.getPureValue()));
  }

  private void addPlannedWeeklyExpense(Payment payment) {
    addDailyValue(payment, getHashtable(plannedWeeklyExpenses));
  }

  public void updatePlannedWeeklyExpense(Payment oldPayment, Payment newPayment) {
    SimpleDate newDate = newPayment.getDate();
    String newName = newPayment.getName();
    long newValue = newPayment.getPureValue();
    SimpleDate oldDate = oldPayment.getDate();
    String oldName = oldPayment.getName();
    long oldValue = oldPayment.getPureValue();
    boolean dateDiff = !newDate.equals(oldDate);
    boolean nameDiff = !newName.equals(oldName);
    if (dateDiff || nameDiff) {
      removePlannedWeeklyExpense(oldPayment);
    }
    if (dateDiff || nameDiff || (newValue != oldValue)) {
      addPlannedWeeklyExpense(newPayment);
    }
  }

  public void removePlannedWeeklyExpense(Payment payment) {
    removeDailyValue(payment, getHashtable(plannedWeeklyExpenses));
  }

  public void updatePlannedDailyExpense(long day, DailyPlanning planning) {
    Long dayLong = new Long(day);
    Hashtable plannedDailyExpenses = getHashtable(this.plannedDailyExpenses);
    plannedDailyExpenses.remove(dayLong);
    if (!DailyPlanningFactory.isRemovable(planning, day)) {
      plannedDailyExpenses.put(dayLong, planning);
    }
  }

  // card income:

  private void addCardIncome(Payment payment) {
    addDailyValue(payment, getHashtable(cardIncome));
  }

  public void updateCardIncome(Payment oldPayment, Payment newPayment) {
    SimpleDate newDate = newPayment.getDate();
    String newName = newPayment.getName();
    long newValue = newPayment.getPureValue();
    SimpleDate oldDate = oldPayment.getDate();
    String oldName = oldPayment.getName();
    long oldValue = oldPayment.getPureValue();
    boolean dateDiff = !newDate.equals(oldDate);
    boolean nameDiff = !newName.equals(oldName);
    if (dateDiff || nameDiff) {
      removeCardIncome(oldPayment);
    }
    if (dateDiff || nameDiff || (newValue != oldValue)) {
      addCardIncome(newPayment);
    }
  }

  public void removeCardIncome(Payment payment) {
    removeDailyValue(payment, getHashtable(cardIncome));
  }

  // card withdraw:

  private void addCardWithdraw(Payment payment) {
    addDailyValue(payment, getHashtable(cardWithdraw));
  }

  public void updateCardWithdraw(Payment oldPayment, Payment newPayment) {
    SimpleDate newDate = newPayment.getDate();
    String newName = newPayment.getName();
    long newValue = newPayment.getPureValue();
    SimpleDate oldDate = oldPayment.getDate();
    String oldName = oldPayment.getName();
    long oldValue = oldPayment.getPureValue();
    boolean dateDiff = !newDate.equals(oldDate);
    boolean nameDiff = !newName.equals(oldName);
    if (dateDiff || nameDiff) {
      removeCardWithdraw(oldPayment);
    }
    if (dateDiff || nameDiff || (newValue != oldValue)) {
      addCardWithdraw(newPayment);
    }
  }

  public void removeCardWithdraw(Payment payment) {
    removeDailyValue(payment, getHashtable(cardWithdraw));
  }

  // balance methods:

  public long getTotalMonthlyIncome(SimpleDate month) {
    return sumMonthlyValues(month, getHashtable(monthlyIncome));
  }

  private long getSumOfPlannedMonthlyExpenses(SimpleDate month) {
    return sumMonthlyValues(month, getHashtable(plannedMontlyExpenses));
  }

  private long getSumOfPlannedDailyExpenses(SimpleDate beginDate, long afterEnd) {
    long begin = beginDate.toLong();
    long plannedEveryDayExpense = getPlannedEveryDayExpense(beginDate);
    Hashtable plannedDailyExpenses = getHashtable(this.plannedDailyExpenses);
    Hashtable plannedWeeklyExpenses = getHashtable(this.plannedWeeklyExpenses);
    Long[] daysOfWeek = new Long[7];
    for (int i = 0; i < 7; i++) {
      daysOfWeek[i] = new Long(beginDate.findDayOfWeekInMonth(i).toLong());
    }
    int dayOfWeek = beginDate.getDayOfWeek();
    long result = 0;
    for (long i = begin; i < afterEnd; i++) {
      DailyPlanning planning = (DailyPlanning)plannedDailyExpenses.get(new Long(i));
      boolean daily;
      boolean weekly;
      long pureValue;
      if (planning == null) {
        planning = DailyPlanningFactory.getRemovable(dayOfWeek);
      }

      daily = planning.isDaily();
      weekly = planning.isWeekly();
      pureValue = planning.getPureValue();

      result += pureValue;
      if (daily) {
        result += plannedEveryDayExpense;
      }
      if (weekly) {
        result += sumHashtable(
          (Hashtable)plannedWeeklyExpenses.get(daysOfWeek[dayOfWeek]));
      }
      dayOfWeek = (dayOfWeek + 1) % 7;
    }
    return result;
  }

  private long getSumOfPlannedDailyExpenses(SimpleDate month) {
    month = month.getMonthAsDate();
    SimpleDate nextMonth = month.getNextMonth();
    long afterEnd = nextMonth.toLong();
    return getSumOfPlannedDailyExpenses(month, afterEnd);
  }

  public long getExpenses(SimpleDate month) {
    return
      getSumOfPlannedMonthlyExpenses(month) +
      getDailySpentInMonth(month);
  }

  public long getDailySpentInMonth(SimpleDate month) {
    return sumDailyValues(month, null, getHashtable(dailyExpenses));
  }

  public long getPlannedExpenses(SimpleDate month) {
    return
      getSumOfPlannedMonthlyExpenses(month) +
      getSumOfPlannedDailyExpenses(month);
  }

  public long getPlannedBalance(SimpleDate month) {
    return getTotalMonthlyIncome(month) - getPlannedExpenses(month);
  }

  public long getCurrentBalance(SimpleDate day) {
    return getTotalMonthlyIncome(day) -
      getSumOfPlannedMonthlyExpenses(day) -
      sumDailyValues(day, day, getHashtable(dailyExpenses)) -
      getSumOfPlannedDailyExpenses(
        new SimpleDate(day.toLong() + 1),
        day.getNextMonth().toLong());
  }

  public long getCurrentPlannedBalance(SimpleDate day) {
    return getCurrentBalance(day) - getPlannedBalance(day);
  }

  public long getTotalDailyPlanning(SimpleDate day) {
    long planned = 0;
    Long dayLong = new Long(day.toLong());
    DailyPlanning planning =
      (DailyPlanning)getHashtable(plannedDailyExpenses).get(
        dayLong);
    boolean daily;
    boolean weekly;
    long pureValue;
    int dayOfWeek = day.getDayOfWeek();
    if (planning == null) {
      planning = DailyPlanningFactory.getRemovable(dayOfWeek);
    }

    daily = planning.isDaily();
    weekly = planning.isWeekly();
    pureValue = planning.getPureValue();

    planned += pureValue;
    if (daily) {
      planned += getPlannedEveryDayExpense(day);
    }
    if (weekly) {
      planned += getPlannedWeeklyExpense(day);
    }
    return planned;
  }

  public long getPlannedWeeklyExpense(SimpleDate day) {
    return sumHashtable(
      (Hashtable)getHashtable(plannedWeeklyExpenses).get(
        new Long(day.getDayOfWeekAsDate().toLong())));
  }

  public long getDailyBalance(SimpleDate day) {
    long planned = getTotalDailyPlanning(day);
    Long dayLong = new Long(day.toLong());
    long real = sumHashtable(
      (Hashtable)getHashtable(dailyExpenses).get(dayLong));
    return planned - real;
  }

  // card money:
  public long getCardMoney(SimpleDate month) {
    return sumDailyValues(month, null, getHashtable(cardIncome)) -
           sumDailyValues(month, null, getHashtable(cardWithdraw));
  }

  public long getCacheMoney(SimpleDate month) {
    return getCurrentMoney(month) -
           getCardMoney(month);
  }

  public long getCurrentMoney(SimpleDate month) {
    return getCurrentBalance(month.getEndOfMonth());
  }

  // display methods:

  private Payment[] ret(Vector result) {
    return (Payment[])result.toArray(new Payment[result.size()]);
  }

  public Payment[] getDailyExpenses(SimpleDate month) {
    return collectDaily(month, getHashtable(dailyExpenses));
  }

  public Payment[] getDailyExpensesSummaries(SimpleDate month) {
    return collectDailySummaries(month, getHashtable(dailyExpenses));
  }

  public Payment[] getMonthlyIncome(SimpleDate month) {
    return collectMonthly(month, getHashtable(monthlyIncome));
  }

  public Payment[] getPlannedMonthlyExpenses(SimpleDate month) {
    return collectMonthly(month, getHashtable(plannedMontlyExpenses));
  }

  public long getPlannedEveryDayExpense(SimpleDate month) {
    Long monthLong = new Long(month.getMonthAsDate().toLong());
    Long everyDayExpensesLong = (Long)getHashtable(plannedEveryDayExpenses).get(monthLong);
    long everyDayExpenses = everyDayExpensesLong != null ?
      everyDayExpensesLong.longValue() : 0;
    return everyDayExpenses;
  }

  public Payment[] getPlannedWeeklyExpenses(SimpleDate month) {
    return collectDaily(month, getHashtable(plannedWeeklyExpenses));
  }

  public Payment[] getPlannedWeeklyExpensesSummaries(SimpleDate month) {
    int day = month.findDayOfWeekInMonth(6).getDay() + 1;
    return collectDailySummaries(
      month.findDayOfWeekInMonth(0),
      new SimpleDate(day, month.getMonth(), month.getYear()),
      getHashtable(plannedWeeklyExpenses));
  }

  public DailyPlanning[] getPlannedDailyExpenses(SimpleDate month) {
    Vector result = new Vector(32);
    month = month.getMonthAsDate();
    SimpleDate nextMonth = month.getNextMonth();
    long begin = month.toLong();
    long afterEnd = nextMonth.toLong();
    Hashtable plannedDailyExpenses = getHashtable(this.plannedDailyExpenses);
    for (long i = begin; i < afterEnd; i++) {
      DailyPlanning dailyPlanning = (DailyPlanning)plannedDailyExpenses.get(new Long(i));
      if (dailyPlanning == null) {
        dailyPlanning = DailyPlanningFactory.getRemovable(
          SimpleDate.getDayOfWeekForLong(i));
      }
      result.addElement(dailyPlanning);
    }
    return (DailyPlanning[])result.toArray(new DailyPlanning[result.size()]);
  }

  public Payment[] getCardIncome(SimpleDate month) {
    return collectDaily(month, getHashtable(cardIncome));
  }

  public Payment[] getCardIncomeSummaries(SimpleDate month) {
    return collectDailySummaries(month, getHashtable(cardIncome));
  }

  public Payment[] getCardWithdraw(SimpleDate month) {
    return collectDaily(month, getHashtable(cardWithdraw));
  }

  public Payment[] getCardWithdrawSummaries(SimpleDate month) {
    return collectDailySummaries(month, getHashtable(cardWithdraw));
  }

}
